/*
 * Isolieren von GPS-Daten aus einem
 * NMEA-Datensrom
 */
#include "string.h"
#include "stdlib.h"
#include "Nmea2Gps.h"

static unsigned long status = 0;

#define NMEA2GPS_STATUS_SYNC 0
#define NMEA2GPS_STATUS_MSG_ID 1
#define NMEA2GPS_STATUS_LATITUDE 2
#define NMEA2GPS_STATUS_LATITUDE_N_S 3
#define NMEA2GPS_STATUS_LONGITUDE 4
#define NMEA2GPS_STATUS_LONGITUDE_E_W 5
#define NMEA2GPS_STATUS_UTC 6
#define NMEA2GPS_STATUS_SKIP_1 7
#define NMEA2GPS_STATUS_SKIP_2 8
#define NMEA2GPS_STATUS_SKIP_3 9
#define NMEA2GPS_STATUS_SKIP_4 10
#define NMEA2GPS_STATUS_SKIP_5 11
#define NMEA2GPS_STATUS_SKIP_6 12
#define NMEA2GPS_STATUS_SATELLITES 13
#define NMEA2GPS_STATUS_SKIP_7 14
#define NMEA2GPS_STATUS_ALTITUDE 15
#define NMEA2GPS_STATUS_COMPLETE 99

/*
 * Die GPS-Daten, die gerade eingelesen werden.
 */
static Nmea2GpsType hotGPS;

/*
 * Die GPS-Daten, die komplett eingelesen wurden.
 */
Nmea2GpsType Nmea2GpsData;

/*
 * Zeichen, das es ein Update gab
 */
unsigned long Nmea2GpsUpdated;

/*
 * Puffer fr empfangene Zeichen
 */
#define MAXBUF 10
static char recvChar[MAXBUF];

/*
 * Index auf den Empfangpuffer
 */
static unsigned long recvCharIdx;

/*
 * gibt an ob ein Paramter kopmlett
 * eingelesen wurde
 */
static unsigned long parameterComplete;

/*
 * Fgt das Zeichen an den Buffer an.
 * Ist das Zeichen in ',' oder ist der Buffer
 * voll, wird true zurckbergeben
 */
void char2Buffer(char ch){

    /*
     * Parameter ende ?
     */
    if ((ch ==',') || recvCharIdx == MAXBUF){

        /*
         * String mit 0x00 abschliessen
         * und Index wieder auf Startposition
         */
        recvChar[recvCharIdx] = 0x00;
        recvCharIdx = 0;
        parameterComplete = 1;
        return;
    }

    /*
     * Normales Zeichen kommt in den Buffer
     */
    recvChar[recvCharIdx] = ch;
    recvCharIdx++;
    parameterComplete = 0;
}

/* * * * * * * * * * * * * * * * * * *
 *    Scanner der einzelnen Phasen
 * * * * * * * * * * * * * * * * * * */

/*
 * Auf den nchsten Message-block synchronisieren
 */
void scan4NMEA2GPS_STATUS_SYNC(){
    /*
     * Wenn das Empfange Zeichen ein $ ist,
     * geht's mit dem Decodieren los.
     * Der Index des Biffers wird immer auf
     * 0 festgesetzt, damit das nchste Zeichen
     * immer in Position 0 landet.
     */
    recvCharIdx = 0;
    if( recvChar[0] == '$'){
        status = NMEA2GPS_STATUS_MSG_ID;
    }
}

/*
 * Auf den nchsten Message-block synchronisieren
 */
void scan4NMEA2GPS_STATUS_MSG_ID(){
    /*
     * $GPGLL gefunden
     */
    if (!strcmp(recvChar,"GPGLL")){
        status = NMEA2GPS_STATUS_LATITUDE;
        recvCharIdx = 0;
        return;
    }

    /*
     * $GPGGA gefunden
     */
    if (!strcmp(recvChar,"GPGGA")){
        status = NMEA2GPS_STATUS_SKIP_1;
        recvCharIdx = 0;
        return;
    }

    /*
     * $GPGGA gefunden => Dies ist der letzte
     * Record eines packetes. Er triggert die
     * bernahme in den Ausgabepuffer.
     */
    if (!strcmp(recvChar,"GPVTG")){
        status = NMEA2GPS_STATUS_COMPLETE;
        recvCharIdx = 0;
        return;
    }

    /*
     * Auf keine passende Message eingerastet
     * -> neu Synchronisieren
     */
    status = NMEA2GPS_STATUS_SYNC;
}

/*
 * Aus dem $GPGLL die Latitude herausholen
 */
void scan4NMEA2GPS_STATUS_LATITUDE(){
    char s[5];

    /*
     * Die ersten zwei Zeichen sind die Grad
     */
    strncpy(s, recvChar, 2);
    s[2]=0x00;
    hotGPS.LatitudeDegree = strtol(s, NULL, 10);

    /*
     * Die nchsten zwei zeichen sin die Minute
     */
    strncpy(s, &recvChar[2], 2);
    s[2]=0x00;
    hotGPS.LatitudeMinuteGrad = strtol(s, NULL, 10);

    /*
     * Zwei Stellen nach dem Punkt sind centi-Minuten
     */
    strncpy(s, &recvChar[5], 2);
    s[2]=0x00;
    hotGPS.LatitudeMinuteCent = strtol(s, NULL, 10);

    /*
     * Der Rest sind 100stel sekunden
     */
    strncpy(s, &recvChar[7], 2);
    s[2]=0x00;
    hotGPS.LatitudeMinuteMyrio = strtol(s, NULL, 10);

    /*
     * Weiter geht's mit Nod-Sd-Kennung
     */
    status = NMEA2GPS_STATUS_LATITUDE_N_S;
}


/*
 * Aus dem $GPGLL die Nord-Sd-Kennung rausholen
 */
void scan4NMEA2GPS_STATUS_LATITUDE_N_S(){
    /*
     * Das Erste Zeichen im Puffer bestimmt die
     * Nord/Sr-Richtung; Nord = + ; Sd = -;
     */
    if (recvChar[0] == 'S'){
        hotGPS.LatitudeDegree *= -1;
    }

    /*
     * Weiter geht's mit Nod-Sd-Kennung
     */
    status = NMEA2GPS_STATUS_LONGITUDE;
}

/*
 * Aus dem $GPGLL die Longitude rausholen
 */
void scan4NMEA2GPS_STATUS_LONGITUDE(){
    char s[5];

    /*
     * Die ersten drei Zeichen sind die Grad
     */
    strncpy(s, recvChar, 3);
    s[3]=0x00;
    hotGPS.LongitudeDegree = strtol(s, NULL, 10);

    /*
     * Die nchsten zwei zeichen sin die Minute
     */
    strncpy(s, &recvChar[3], 2);
    s[2]=0x00;
    hotGPS.LongitudeMinuteGrad = strtol(s, NULL, 10);

    /*
     * Zwei Stellen nach dem Punkt sind 100stel Minuten
     */
    strncpy(s, &recvChar[6], 2);
    s[2]=0x00;
    hotGPS.LongitudeMinuteCent = strtol(s, NULL, 10);

    /*
     * Zwei Stellen danch 10000stel Minuten
     */
    strncpy(s, &recvChar[8], 2);
    s[2]=0x00;
    hotGPS.LongitudeMinuteMyrio = strtol(s, NULL, 10);

    /*
     * Weiter geht's mit Ost-West-Kennung
     */
    status = NMEA2GPS_STATUS_LONGITUDE_E_W;

}

/*
 * Aus dem $GPGLL die Ost-West-Kennung rausholen
 */
void scan4NMEA2GPS_STATUS_LONGITUDE_E_W(){
    /*
     * Das Erste Zeichen im Puffer bestimmt die
     * Nord/Sr-Richtung; Ost = + ; West = -;
     */
    if (recvChar[0] == 'W'){
        hotGPS.LongitudeDegree *= -1;
    }

    /*
     * Weiter geht's mit Nod-Sd-Kennung
     */
    status = NMEA2GPS_STATUS_UTC;
}


/*
 * Aus dem $GPGLL die Zeit rausholen
 */
void scan4NMEA2GPS_STATUS_UTC(){
    char s[4];

    /*
     * Die ersten zwei Zeichen sind die Stunde
     */
    strncpy(s, recvChar, 2);
    s[2]=0x00;
    hotGPS.UtcHour = strtol(s, NULL, 10);

    /*
     * Die nchsten zwei zeichen sin die Minute
     */
    strncpy(s, &recvChar[2], 2);
    s[2]=0x00;
    hotGPS.UtcMinute = strtol(s, NULL, 10);

    /*
     * Die nchsten zwei Zeichen sind die Sekunde
     */
    strncpy(s, &recvChar[4], 2);
    s[2]=0x00;
    hotGPS.UtcSecunde = strtol(s, NULL, 10);

    /*
     * Dieser Datensatz ist komplett. Es wird auf
     * den nchsten Datensatz geweartet.
     */
    status = NMEA2GPS_STATUS_SYNC;
}


/*
 * Parameter UTC-Time im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_1(){
    status = NMEA2GPS_STATUS_SKIP_2;
}

/*
 * Parameter Latitude im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_2(){
    status = NMEA2GPS_STATUS_SKIP_3;
}

/*
 * Parameter N/S im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_3(){
    status = NMEA2GPS_STATUS_SKIP_4;
}

/*
 * Parameter Logitude im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_4(){
    status = NMEA2GPS_STATUS_SKIP_5;
}

/*
 * Parameter E/W im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_5(){
    status = NMEA2GPS_STATUS_SKIP_6;
}

/*
 * Parameter Position Fix im Record $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_6(){
    status = NMEA2GPS_STATUS_SATELLITES;
}

/*
 * Aus dem $GPGGA die Anzahl der verwendeten Satelliten ermitteln
 */
void scan4NMEA2GPS_STATUS_SATELLITES(){

    hotGPS.SatellitsInUse = strtol(recvChar, NULL, 10);

    /*
     * Den ncsten Parameter wieder bersppringen
     */
    status = NMEA2GPS_STATUS_SKIP_7;

}

/*
 * Parameter Horizontale Abweichung im $GPGGA berspringen
 */
void scan4NMEA2GPS_STATUS_SKIP_7(){
    status = NMEA2GPS_STATUS_ALTITUDE;
}

/*
 * Aus dem $GPGGA die Anzahl der verwendeten Satelliten ermitteln
 */
void scan4NMEA2GPS_STATUS_ALTITUDE(){

    hotGPS.AltitudeMeter = strtol(recvChar, NULL, 10);

    /*
     * Den nchsten Record suchen
     */
    status = NMEA2GPS_STATUS_SYNC;
}

/*
 * Die GPSdaten sind alle eingesammelt und
 * werden in den Ausgabepuffer kopiert. Im Anschluss
 * wird das Update-Kennzeichen gesetzt.
 */
void scan4NMEA2GPS_STATUS_COMPLETE(){

    Nmea2GpsData = hotGPS;
    Nmea2GpsUpdated = 1;
    /*
     * Den nchsten Record suchen
     */
    status = NMEA2GPS_STATUS_SYNC;
}

void Nmea2GpsNextChar(unsigned char ch){
    /*
     * Zeichen aufpuffern
     */
    char2Buffer(ch);

    /*
     * Status NMEA2GPS_STATUS_SYNC wartet nur
     * auf ein Dollarzeichen. Er kann deshalb sofort
     * arbeiten....
     */
    if (status == NMEA2GPS_STATUS_SYNC){
        scan4NMEA2GPS_STATUS_SYNC();
        return;
    }

    /*
     * ...alle anderen Status werden erst bentigt,
     * wenn ein Parameter komplett eingelesen ist.
     */
    if (!parameterComplete) return;

    /*
     * Paramter ist komplett
     * => Es kann ausgewertete werden.
     */
    switch ( status ){
        /* GPGLL Datensatz */
        case NMEA2GPS_STATUS_MSG_ID:
            scan4NMEA2GPS_STATUS_MSG_ID(); break;

        case NMEA2GPS_STATUS_LATITUDE:
            scan4NMEA2GPS_STATUS_LATITUDE(); break;

        case NMEA2GPS_STATUS_LATITUDE_N_S:
            scan4NMEA2GPS_STATUS_LATITUDE_N_S(); break;

        case NMEA2GPS_STATUS_LONGITUDE:
            scan4NMEA2GPS_STATUS_LONGITUDE(); break;

        case NMEA2GPS_STATUS_LONGITUDE_E_W:
            scan4NMEA2GPS_STATUS_LONGITUDE_E_W(); break;

        case NMEA2GPS_STATUS_UTC:
            scan4NMEA2GPS_STATUS_UTC(); break;

        /* GPGGA Datensatz */
        case NMEA2GPS_STATUS_SKIP_1:
            scan4NMEA2GPS_STATUS_SKIP_1(); break;

        case NMEA2GPS_STATUS_SKIP_2:
            scan4NMEA2GPS_STATUS_SKIP_2(); break;

        case NMEA2GPS_STATUS_SKIP_3:
            scan4NMEA2GPS_STATUS_SKIP_3(); break;

        case NMEA2GPS_STATUS_SKIP_4:
            scan4NMEA2GPS_STATUS_SKIP_4(); break;

        case NMEA2GPS_STATUS_SKIP_5:
            scan4NMEA2GPS_STATUS_SKIP_5(); break;

        case NMEA2GPS_STATUS_SKIP_6:
            scan4NMEA2GPS_STATUS_SKIP_6(); break;

        case NMEA2GPS_STATUS_SATELLITES:
            scan4NMEA2GPS_STATUS_SATELLITES(); break;

        case NMEA2GPS_STATUS_SKIP_7:
            scan4NMEA2GPS_STATUS_SKIP_7(); break;

        case NMEA2GPS_STATUS_ALTITUDE:
            scan4NMEA2GPS_STATUS_ALTITUDE(); break;

        /* GPVTG Datensatz */
        case NMEA2GPS_STATUS_COMPLETE:
            scan4NMEA2GPS_STATUS_COMPLETE(); break;
    }
    return;
}
